import numpy as np
import time

import tensorflow as tf
import tensorflow_probability as tfp
from sccoda.model.dirichlet_models import CompositionalModel

tfd = tfp.distributions
tfb = tfp.bijectors

#%%


class NoBaselineModelTime(CompositionalModel):

    """
    WIP! DO NOT USE!!!

    implements statistical model and
    test statistics for compositional differential change analysis
    without specification of a baseline cell type
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor of model class

        :param covariate_matrix: numpy array [NxD] - covariate matrix
        :param data_matrix: numpy array [NxK] - cell count matrix
        :param sample_counts: numpy array [N] - number of cells per sample
        :param dtype: data type for all numbers (for tensorflow)

        :return
        NotImplementedError
        """

        super(self.__class__, self).__init__(*args, **kwargs)

        self.baseline_index = None
        dtype = tf.float32

        # All parameters that are returned for analysis
        self.param_names = ["mu_b", "sigma_b", "b_offset", "ind_raw", "alpha", "phi",
                            "ind", "b_raw", "beta"]

        alpha_size = [self.K]
        beta_size = [self.D, self.K]

        Root = tfd.JointDistributionCoroutine.Root

        def model():
            mu_b = yield Root(tfd.Independent(
                tfd.Normal(loc=tf.zeros(1, dtype=dtype),
                           scale=tf.ones(1, dtype=dtype),
                           name="mu_b"),
                reinterpreted_batch_ndims=1))

            sigma_b = yield Root(tfd.Independent(
                tfd.HalfCauchy(tf.zeros(1, dtype=dtype),
                               tf.ones(1, dtype=dtype),
                               name="sigma_b"),
                reinterpreted_batch_ndims=1))

            b_offset = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(beta_size, dtype=dtype),
                    scale=tf.ones(beta_size, dtype=dtype),
                    name="b_offset"),
                reinterpreted_batch_ndims=2))

            # Spike-and-slab
            ind_raw = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(shape=beta_size, dtype=dtype),
                    scale=tf.ones(shape=beta_size, dtype=dtype),
                    name='ind_raw'),
                reinterpreted_batch_ndims=2))

            ind = tf.exp(ind_raw * 50) / (1 + tf.exp(ind_raw * 50))
            b_raw = mu_b + sigma_b * b_offset
            beta = ind * b_raw

            alpha = yield Root(tfd.Independent(
                tfd.Normal(
                    loc=tf.zeros(alpha_size),
                    scale=tf.ones(alpha_size) * 5,
                    name="alpha"),
                reinterpreted_batch_ndims=1))

            phi = yield Root(tfd.Independent(
                tfd.Normal(tf.ones(beta_size, dtype=dtype),
                           tf.ones(beta_size, dtype=dtype) * 0.1,
                           name="phi"),
                reinterpreted_batch_ndims=2))

            phi_ = tf.repeat(phi[tf.newaxis, :], self.N, axis=0)
            b_time = tf.pow(phi_, self.time_matrix[:, tf.newaxis, tf.newaxis])
            b_ = beta[tf.newaxis, :, :] * b_time

            c_ = self.x[:, :, tf.newaxis] * b_

            concentrations = tf.exp(alpha + tf.reduce_sum(c_, axis=1))

            # Cell count prediction via DirMult
            predictions = yield Root(tfd.Independent(
                tfd.DirichletMultinomial(
                    total_count=tf.cast(self.n_total, dtype),
                    concentration=concentrations,
                    name="predictions"),
                reinterpreted_batch_ndims=1))

        self.model_struct = tfd.JointDistributionCoroutine(model)

        # Joint posterior distribution
        self.target_log_prob_fn = lambda *args:\
            self.model_struct.log_prob(list(args) + [tf.cast(self.y, dtype)])

        # MCMC starting values
        self.params = [tf.zeros(1, name="init_mu_b", dtype=dtype),
                       tf.ones(1, name="init_sigma_b", dtype=dtype),
                       tf.zeros(beta_size, name='init_b_offset', dtype=dtype),
                       tf.zeros(beta_size, name='init_ind_raw', dtype=dtype),
                       tf.zeros(alpha_size, name='init_alpha', dtype=dtype),
                       tf.ones(beta_size, name="init_phi", dtype=dtype),
                       ]

        # self.vars = [tf.Variable(v, trainable=True) for v in self.params]

    # Calculate predicted cell counts (for analysis purposes)
    def get_y_hat(self, states_burnin, num_results, n_burnin):
        chain_size_y = [num_results - n_burnin, self.N, self.K]

        alphas = states_burnin[4]
        alphas_final = alphas.mean(axis=0)

        ind_raw = states_burnin[3] * 50
        mu_b = states_burnin[0]
        sigma_b = states_burnin[1]
        b_offset = states_burnin[2]
        phi = states_burnin[5]

        ind_ = np.exp(ind_raw) / (1 + np.exp(ind_raw))

        b_raw_ = mu_b.reshape((num_results - n_burnin, 1, 1)) + np.einsum("...jk, ...j->...jk", b_offset, sigma_b)

        # phi_ = np.repeat(phi[:, np.newaxis, :, :], self.N, axis=1)

        # print(phi_.shape)

        beta_ = np.einsum("..., ...", ind_, b_raw_)

        # conc_ = np.exp(np.einsum("jk, ...kl->...jl", self.x, beta_)
        #                + alphas.reshape((num_results - n_burnin, 1, self.K)))

        # predictions_ = np.zeros(chain_size_y)
        # for i in range(num_results - n_burnin):
        #     pred = tfd.DirichletMultinomial(self.n_total, conc_[i, :, :]).mean().numpy()
        #     predictions_[i, :, :] = pred

        betas_final = beta_.mean(axis=0)
        states_burnin.append(ind_)
        states_burnin.append(b_raw_)
        # states_burnin.append(b_time_)
        states_burnin.append(beta_)
        # states_burnin.append(conc_)
        # states_burnin.append(predictions_)

        # concentration = np.exp(np.matmul(self.x, betas_final) + alphas_final).astype(np.float32)

        #  y_mean = concentration / np.sum(concentration, axis=1, keepdims=True) * self.n_total.numpy()[:, np.newaxis]
        # print(y_mean)

        return None


"""
Helper functions to convert single-cell data to scCODA compositional analysis data

:authors: Johannes Ostner
"""
import pandas as pd
import anndata as ad
import os


def from_scanpy(adata, cell_type_identifier, covariate_key):
    """
    Converts a single scRNA-seq data set from scanpy to a row of a cell count matrix.

    It is assumed that a column of adata.obs contains the cell type assignment, while covariates (control/disease group, ...) are stored as a subdict in adata.uns

    Usage: cell_counts, covs = from_scanpy(adata, cell_type_identifier="Louvain", covariate_key="covariates")

    Parameters
    ----------
    adata -- anndata object
        single-cell data object from scanpy
    cell_type_identifier -- str
        column name in adata.obs that specifies the cell types
    covariate_key -- str
        key for adata.uns, where the covariate values are stored

    Returns
    -------
    cell_counts -- Numpy array
        cell count vector
    covs -- list
        covariate vector
    """

    # Calculate cell counts for the sample
    cell_counts = adata.obs[cell_type_identifier].value_counts()

    # extracting covariates from uns
    covs = adata.uns[covariate_key]

    return cell_counts, covs


def from_scanpy_list(samples, cell_type_identifier, covariate_key):
    """
    Creates a compositional analysis data set from a list of scanpy data sets.

    To use this function, all data sets need to have one common column in adata.obs that contans the cell type assignment.
    Also, the covariates need to be stored under the same key in adata.uns

    Usage: data = from_scanpy([adata1, adata2, adata3], cell_type_identifier="Louvain", covariate_key="covariates")

    Parameters
    ----------
    samples -- list
        list of scanpy data sets
    cell_type_identifier -- str
        column name in adata.obs that specifies the cell types
    covariate_key -- str
        key for adata.uns, where the covariate values are stored

    Returns
    -------
    data -- CompositionalData object
        A compositional analysis data set
    """

    count_data = pd.DataFrame()
    covariate_data = pd.DataFrame()

    # iterate over anndata objects for each sample
    for s in samples:

        cell_counts, covs = from_scanpy(s, cell_type_identifier, covariate_key)
        count_data = count_data.append(cell_counts, ignore_index=True)
        covariate_data = covariate_data.append(pd.Series(covs), ignore_index=True)

    # Replace NaNs
    count_data = count_data.fillna(0)
    covariate_data = covariate_data.fillna(0)

    return ad.AnnData(X=count_data.values,
                      var=count_data.sum(axis=0).rename("n_cells").to_frame(),
                      obs=covariate_data)


def from_scanpy_dir(path, cell_type_identifier, covariate_key):
    """
    Creates a compositional analysis data set from all scanpy data sets in a directory.

    To use this function, all data sets need to have one common column in adata.obs that contans the cell type assignment.
    Also, the covariates need to be stored under the same key in adata.uns

    Usage: data = from_scanpy("./path/to/directory", cell_type_identifier="Louvain", covariate_key="covariates")

    Parameters
    ----------
    path -- str
        path to directory
    cell_type_identifier -- str
        column name in adata.obs that specifies the cell types
    covariate_key -- str
        key for adata.uns, where the covariate values are stored

    Returns
    -------
    data -- CompositionalData object
        A compositional analysis data set
    """

    count_data = pd.DataFrame()
    covariate_data = pd.DataFrame()

    filenames = os.listdir(path)
    for f in filenames:
        adata = ad.read_h5ad(f)

        cell_counts, covs = from_scanpy(adata, cell_type_identifier, covariate_key)
        count_data = count_data.append(cell_counts, ignore_index=True)
        covariate_data = covariate_data.append(pd.Series(covs), ignore_index=True)

    # Replace NaNs
    count_data = count_data.fillna(0)
    covariate_data = covariate_data.fillna(0)

    return ad.AnnData(X=count_data.values,
                      var=count_data.sum(axis=0).rename("n_cells").to_frame(),
                      obs=covariate_data)


def from_pandas(df, covariate_columns):
    """
    Converts a Pandas DataFrame into a compositional analysis data set.
    The DataFrame must contain one row per sample, columns can be cell types or covariates

    Note that all columns that are not specified as covariates are assumed to be cell counts.

    Usage: data = from_pandas(df, covariate_columns=["cov1", "cov2"])

    Parameters
    ----------
    df -- DataFrame
        A pandas DataFrame with each row representing a sample; the columns can be cell counts or covariates
    covariate_columns -- List
        List of column names that are interpreted as covariates; all other columns will be seen as cell types

    Returns
    -------
    data -- CompositionalData object
        A compositional analysis data set
    """

    covariate_data = df.loc[:, covariate_columns]
    count_data = df.loc[:, ~df.columns.isin(covariate_data)]
    celltypes = pd.DataFrame(index=count_data.columns)

    return ad.AnnData(X=count_data.values,
                      var=celltypes,
                      obs=covariate_data)



"""
This script contains methods to visualize compositional data that was imported into scCODA's data format

:authors: Johannes Ostner
"""

# Setup

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.pyplot import cm

sns.set_style("ticks")


def plot_one_stackbar(y, type_names, title, level_names):

    """
    Plots a stacked barplot for one (discrete) covariate
    Typical use: plot_one_stackbar(data.X, data.var.index, "xyz", data.obs.index)

    Parameters
    ----------
    y -- numpy array
        The count data, collapsed onto the level of interest. i.e. a binary covariate has two rows containing the count
        mean of all samples of each cell type
    type_names -- list-like
        The names of all cell types
    title -- string
        Plot title, usually the covariate's name
    level_names -- list-like
        names of the covariate's levels

    Returns
    -------
    a plot
    """

    plt.figure(figsize=(20, 10))
    n_samples, n_types = y.shape
    r = np.array(range(n_samples))
    sample_sums = np.sum(y, axis=1)
    barwidth = 0.85
    cum_bars = np.zeros(n_samples)
    colors = cm.tab20

    for n in range(n_types):
        bars = [i / j * 100 for i, j in zip([y[k][n] for k in range(n_samples)], sample_sums)]
        plt.bar(r, bars, bottom=cum_bars, color=colors(n % 20), width=barwidth, label=type_names[n])
        cum_bars += bars

    plt.title(title)
    plt.legend(loc='upper left', bbox_to_anchor=(1, 1), ncol=1)
    plt.xticks(r, level_names, rotation=45)

    plt.show()


def plot_feature_stackbars(data, features):

    """
    Plots stackbars for all listed covariates
    Usage: plot_feature_stackbars(data, ["cov1", "cov2", "cov3"])

    Parameters
    ----------
    data -- AnnData object
        A scCODA-compatible data object
    features -- list
        List of all covariates to plot. Specifying "samples" as an element will plot a stackbar with all samples

    Returns
    -------
    One plot for every category

    """

    type_names = data.var.index
    for f in features:

        if f == "samples":
            plot_one_stackbar(data.X, data.var.index, "samples", data.obs.index)
        else:
            levels = pd.unique(data.obs[f])
            n_levels = len(levels)
            feature_totals = np.zeros([n_levels, data.X.shape[1]])

            for level in range(n_levels):
                l_indices = np.where(data.obs[f] == levels[level])
                feature_totals[level] = np.sum(data.X[l_indices], axis=0)

            plot_one_stackbar(feature_totals, type_names=type_names, level_names=levels, title=f)


def grouped_boxplot(data, feature, log_scale=False, *args, **kwargs):
    """
    Grouped boxplot -cell types on x-Axis; one boxplot per feature level for each cell type
    Parameters
    ----------
    data -- AnnData object
        A scCODA-compatible data object
    feature -- string
        The name of the feature in data.obs to plot
    log_scale -- bool
        If true, use log(data.X + 1) (pseudocount 1 to avoid log(0)-issues)
    *args, **kwargs -- Passed to sns.boxplot

    Returns
    -------
    Plot!
    """

    plt.figure(figsize=(20, 10))

    # add pseudocount 1 if using log scale (needs to be improved)
    if log_scale:
        X = np.log(data.X + 1)
        value_name = "log(count)"
    else:
        X = data.X
        value_name = "count"

    count_df = pd.DataFrame(X, columns=data.var.index, index=data.obs.index).\
        merge(data.obs[feature], left_index=True, right_index=True)
    plot_df = pd.melt(count_df, id_vars=feature, var_name="Cell type", value_name=value_name)

    d = sns.boxplot(x="Cell type", y=value_name, hue=feature, data=plot_df, fliersize=1,
                    palette='Blues', *args, **kwargs)

    loc, labels = plt.xticks()
    d.set_xticklabels(labels, rotation=90)

    return d


def boxplot_facets(data, feature, log_scale=False, args_boxplot={}, args_swarmplot={}):
    """
    Grouped boxplot -cell types on x-Axis; one boxplot per feature level for each cell type
    Parameters
    ----------
    data -- AnnData object
        A scCODA-compatible data object
    feature -- string
        The name of the feature in data.obs to plot
    log_scale -- bool
        If true, use log(data.X + 1) (pseudocount 1 to avoid log(0)-issues)
    args_boxplot -- dict
        Arguments passed to sns.boxplot
    args_swarmplot -- dict
        Arguments passed to sns.swarmplot

    Returns
    -------
    Plot!
    """

    # add pseudocount 1 if using log scale (needs to be improved)
    if log_scale:
        X = np.log(data.X + 1)
        value_name = "log(count)"
    else:
        X = data.X
        value_name = "count"

    K = X.shape[1]

    count_df = pd.DataFrame(X, columns=data.var.index, index=data.obs.index).merge(data.obs, left_index=True,
                                                                                   right_index=True)
    plot_df = pd.melt(count_df, id_vars=data.obs.columns, var_name="Cell type", value_name=value_name)

    if "hue" in args_swarmplot:
        hue = args_swarmplot.pop("hue")
    else:
        hue = None

    g = sns.FacetGrid(plot_df, col="Cell type", sharey=False, col_wrap=np.floor(np.sqrt(K)), height=5, aspect=2)
    g.map(sns.boxplot, feature, value_name, palette="Blues", **args_boxplot)
    if hue is None:
        g.map(sns.swarmplot, feature, value_name, color="black", **args_swarmplot).set_titles("{col_name}")
    else:
        g.map(sns.swarmplot, feature, value_name, hue, **args_swarmplot).set_titles("{col_name}")
    return g
